#!/usr/bin/python

r'''Constructs README, README.org files

The README files are generated by this script. They are made from:

- The main module docstring, with some org markup applied to the README.org, but
  not to the README
- The docstrings from each API function in the module, with some org markup
  applied to the README.org, but not to the README
- README.footer.org, copied verbatim

The main module name must be passed in as the first cmdline argument.
'''

import sys

try:
    modname,readmeorg,readme,readmefooterorg = sys.argv[1:]
except:
    raise Exception("Usage: {} modulename readmeorg readme readmefooterorg".format(sys.argv[0]))

exec( 'import {} as mod'.format(modname) )

import inspect
import re
try:
    from StringIO import StringIO ## for Python 2
except ImportError:
    from io import StringIO ## for Python 3

def dirmod():
    r'''Same as dir(mod), but returns only functions, in the definition order'''
    with open('{}.py'.format(modname), 'r') as f:
        for l in f:
            m = re.match(r'def +([a-zA-Z0-9_]+)\(', l)
            if m:
                yield m.group(1)

with open(readmeorg, 'w') as f_target_org:
    with open(readme, 'w') as f_target:

        f_target_org.write(r'''* TALK
I just gave a talk about this at [[https://www.socallinuxexpo.org/scale/18x][SCaLE 18x]]. Presentation lives [[https://github.com/dkogan/talk-numpysane-gnuplotlib/raw/master/numpysane-gnuplotlib.pdf][here]].

''')

        def write(s):
            f_target.    write(s)
            f_target_org.write(s)

        def write_orgized(s):
            r'''Writes the given string, reformatted slightly with org in mind.

            The only change this function applies, is to look for indented block
            (signifying examples) and to wrap then in a #+BEGIN_SRC or
            #+BEGIN_EXAMPLE.

            '''

            # the non-org version is written as is
            f_target.write(s)

            # the org version neeeds massaging
            f = f_target_org

            in_quote = False
            queued_blanks = 0
            indent_size = 4

            prev_indented = False

            sio = StringIO(s)
            for l in sio:

                if not in_quote:
                    if len(l) <= 1:
                        # blank line
                        f.write(l)
                        continue

                    if not re.match(' '*indent_size, l):
                        # don't have full indent. not quote start
                        prev_indented = re.match(' ', l)
                        f.write(l)
                        continue

                    if re.match(' '*indent_size + '-', l):
                        # Start of indented list. not quote start
                        prev_indented = re.match(' ', l)
                        f.write(l)
                        continue

                    if prev_indented:
                        # prev line(s) were indented, so this can't start a quote
                        f.write(l)
                        continue

                    # start of quote. What kind?
                    in_quote = True
                    f.write('#+BEGIN_EXAMPLE\n')

                    f.write(l[indent_size:])
                    continue

                # we're in a quote. Skip blank lines for now
                if len(l) <= 1:
                    queued_blanks = queued_blanks+1
                    continue

                if re.match(' '*indent_size, l):
                    # still in quote. Write it out
                    f.write( '\n'*queued_blanks)
                    queued_blanks = 0
                    f.write(l[indent_size:])
                    continue

                # not in quote anymore
                f.write('#+END_EXAMPLE\n')
                f.write( '\n'*queued_blanks)
                f.write(l)
                queued_blanks = 0
                in_quote      = False
                prev_indented = False

            f.write('\n')
            if in_quote: f.write('#+END_EXAMPLE\n')




        header = '* NAME\n{}: '.format(modname)
        write( header )






        write_orgized(inspect.getdoc(mod))
        write( '\n' )

        def getdoc(func):
            if re.match('_', func):
                return None
            if not inspect.isfunction(mod.__dict__[func]):
                return None
            doc = inspect.getdoc(mod.__dict__[func])
            if not doc:
                return None
            return doc
        funcs_docs = [(f,getdoc(f)) for f in dirmod()]
        funcs_docs = [(f,d) for f,d in funcs_docs if d is not None]

        if len(funcs_docs):
            write('* INTERFACE\n')

            for func,doc in funcs_docs:
                write('** {}()\n'.format(func))
                write_orgized( doc )
                write( '\n' )

        with open(readmefooterorg, 'r') as f_footer:
            write( f_footer.read() )



#!/usr/bin/python3

r'''generate broadcast-aware python wrapping to testlib

The test suite runs this script to python-wrap the testlib C library, then
the test suite builds this python extension module, and then the test suite
validates this module's behavior

'''
import sys
import os

dir_path = os.path.dirname(os.path.realpath(__file__))
sys.path[:0] = dir_path + '/..',

import numpy as np
import numpysane as nps
import numpysane_pywrap as npsp


docstring_module = r"""Some functions to test the python wrapping

multiline docstring to stress-test this thing
line
line
line
"""
m = npsp.module( name      = "testlib",
                 docstring = docstring_module,
                 header    = '''
#include <stdlib.h>
#include "testlib.h"
''')

m.function( "identity3",
            r"""Generates a 3x3 identity matrix

multi-line docstring to make sure it works
""",

            args_input       = (),
            prototype_input  = (),
            prototype_output = (3,3),

            Ccode_slice_eval = \
                {np.float64:
                 r'''
    for(int i=0; i<3; i++)
        for(int j=0; j<3; j++)
            item__output(i,j) = (i==j) ? 1.0 : 0.0;
    return true;
'''})

m.function( "identity",
            '''Generates an NxN identity matrix. Output matrices must be passed-in to define N''',

            args_input       = (),
            prototype_input  = (),
            prototype_output = ('N', 'N'),

            Ccode_slice_eval = \
                {np.float64:
                 r'''
    int N = dims_slice__output[0];
    for(int i=0; i<N; i++)
        for(int j=0; j<N; j++)
            item__output(i,j) = (i==j) ? 1.0 : 0.0;
    return true;
'''},)

m.function( "inner",
            "Inner-product pywrapped with npsp",

            args_input       = ('a', 'b'),
            prototype_input  = (('n',), ('n',)),
            prototype_output = (),

            Ccode_slice_eval = \
                {np.float64:
                 r'''
            const int N = dims_slice__a[0];
            item__output() =
              inner_double((double*)data_slice__a,
                           (double*)data_slice__b,
                           strides_slice__a[0],
                           strides_slice__b[0],
                           N);
            return true;
''',
                 np.int64:
                 r'''
            const int N = dims_slice__a[0];
            item__output() =
              inner_int64_t((int64_t*)data_slice__a,
                            (int64_t*)data_slice__b,
                            strides_slice__a[0],
                            strides_slice__b[0],
                            N);
            return true;
''',
                 np.int32:
                 r'''
            const int N = dims_slice__a[0];
            item__output() =
              inner_int32_t((int32_t*)data_slice__a,
                            (int32_t*)data_slice__b,
                            strides_slice__a[0],
                            strides_slice__b[0],
                            N);
            return true;
'''})

m.function( "outer",
            "Outer-product pywrapped with npsp",

            args_input       = ('a', 'b'),
            prototype_input  = (('n',), ('n',)),
            prototype_output = ('n', 'n'),

            Ccode_slice_eval = \
                {np.float64:
                 r'''
            int N = dims_slice__a[0];
            outer((double*)data_slice__output,
                  strides_slice__output[0],
                  strides_slice__output[1],
                  (double*)data_slice__a,
                  (double*)data_slice__b,
                  strides_slice__a[0],
                  strides_slice__b[0],
                  N);
            return true;
'''})

m.function( "innerouter",
            "Inner and outer products pywrapped with npsp",

            args_input       = ('a', 'b'),
            prototype_input  = (('n',), ('n',)),
            prototype_output = ((), ('n', 'n')),

            Ccode_cookie_struct = '''
            double scale; /* from BOTH scale arguments: "scale", "scale_string" */
            char*  ptr;   /* to demo resource allocation/release */
            ''',

            extra_args = (("double",      "scale",          "1",    "d"),
                          ("const char*", "scale_string",   "NULL", "s")),

            Ccode_validate = r'''
            if(! (*scale > 0.0       &&
                  CHECK_CONTIGUOUS_AND_SETERROR_ALL() ) )
                return false;
            cookie->scale = *scale * (scale_string ? atof(scale_string) : 1.0);
            cookie->ptr   = malloc(1000);
            return cookie->ptr != NULL;
''',

            Ccode_slice_eval = \
                {np.float64:
                 r'''
            int N = dims_slice__a[0];
            item__output0() =
                 innerouter((double*)data_slice__output1,
                     (double*)data_slice__a,
                     (double*)data_slice__b,
                     cookie->scale,
                     N);
            return true;
'''},
            Ccode_cookie_cleanup = 'if(cookie->ptr != NULL) free(cookie->ptr);'
           )

m.function( "sorted_indices",
            "Return the sorted element indices",

            args_input       = ('x',),
            prototype_input  = (('n',),),
            prototype_output = ('n',),

            Ccode_slice_eval = \
                {(np.float32, np.int32):
                 r'''
                 sorted_indices_float((int*)data_slice__output,
                     (float*)data_slice__x,
                     dims_slice__x[0]);
                 return true;
''',

                 (np.float64, np.int32):
                 r'''
                 sorted_indices_double((int*)data_slice__output,
                     (double*)data_slice__x,
                     dims_slice__x[0]);
                 return true;
'''},
            Ccode_validate = 'return CHECK_CONTIGUOUS_AND_SETERROR__x();' )

# Tests. Try to wrap functions using illegal output prototypes. The wrapper code
# should barf

try:
    m.function( "outer_broken",
                "Outer-product pywrapped with npsp",
                args_input       = ('a', 'b'),
                prototype_input  = (('n',), ('n',)),
                prototype_output = ('n', -1),
                Ccode_slice_eval = {np.float64: 'return true;'})
except: pass # known error
else:   raise Exception("Expected error didn't happen")

try:
    m.function( "outer_broken",
                "Outer-product pywrapped with npsp",
                args_input       = ('a', 'b'),
                prototype_input  = (('n',), (-1,)),
                prototype_output = ('n', 'n'),
                Ccode_slice_eval = {np.float64: 'return true;'})
except: pass # known error
else:   raise Exception("Expected error didn't happen")

try:
    m.function( "outer_broken",
                "Outer-product pywrapped with npsp",
                args_input       = ('a', 'b'),
                prototype_input  = (('n',), ('n',-1)),
                prototype_output = (),
                Ccode_slice_eval = {np.float64: 'return true;'})
except: pass # known error
else:   raise Exception("Expected error didn't happen: input dims must >=0")

try:
    m.function( "outer_broken",
                "Outer-product pywrapped with npsp",
                args_input       = ('a', 'b'),
                prototype_input  = (('n',), ('n',)),
                prototype_output = (-1,),
                Ccode_slice_eval = {np.float64: 'return true;'})
except: pass # known error
else:   raise Exception("Expected error didn't happen: output dims must >=0")

try:
    m.function( "outer_broken",
                "Outer-product pywrapped with npsp",
                args_input       = ('a', 'b'),
                prototype_input  = (('n',), ('n',)),
                prototype_output = ('m', ()),
                Ccode_slice_eval = {np.float64: 'return true;'})
except: pass # known error
else:   raise Exception("Expected error didn't happen: output dims must be integers or strings")

try:
    m.function( "outer_broken",
                "Outer-product pywrapped with npsp",
                args_input       = ('a', 'b'),
                prototype_input  = (('n',), ('n',)),
                prototype_output = 'n',
                Ccode_slice_eval = {np.float64: 'return true;'})
except: pass # known error
else:   raise Exception("Expected error didn't happen: output dims must be a tuple")

try:
    m.function( "outer_broken",
                "Outer-product pywrapped with npsp",
                args_input       = ('a', 'b'),
                prototype_input  = (('n',), 'n'),
                prototype_output = 'n',
                Ccode_slice_eval = {np.float64: 'return true;'})
except: pass # known error
else:   raise Exception("Expected error didn't happen: output dims must be a tuple")

try:
    m.function( "sorted_indices_broken",
                "Return the sorted element indices",
                args_input       = ('x',),
                prototype_input  = (('n',),),
                prototype_output = ('n',),
                Ccode_slice_eval = { np.float64: 'return true;' })
except: raise Exception("Valid usage of Ccode_slice_eval keys failed")

try:
    m.function( "sorted_indices_broken2",
                "Return the sorted element indices",
                args_input       = ('x',),
                prototype_input  = (('n',),),
                prototype_output = ('n',),
                Ccode_slice_eval = { np.float64: 'return true;', np.int32: 'return true;' })
except: raise Exception("Valid usage of Ccode_slice_eval keys failed")

try:
    m.function( "sorted_indices_broken3",
                "Return the sorted element indices",
                args_input       = ('x',),
                prototype_input  = (('n',),),
                prototype_output = ('n',),
                Ccode_slice_eval = { (np.float64, np.int32): 'return true;', np.int32: 'return true;' })
except: raise Exception("Valid usage of Ccode_slice_eval keys failed")

try:
    m.function( "sorted_indices_broken4",
                "Return the sorted element indices",
                args_input       = ('x',),
                prototype_input  = (('n',),),
                prototype_output = ('n',),
                Ccode_slice_eval = { (np.float64, np.int32, np.int32): 'return true;', np.int32: 'return true;' })
except: pass # known error
else:   raise Exception("Expected invalid usage of Ccode_slice_eval keys didn't fail!")


m.write()
